/*
    Ming, an SWF output library
    Copyright (C) 2000  Opaque Industries - http://www.opaque.net/

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <math.h>

#include "shape.h"

/* draw an arc of radius r, centered at (x,y), from angle startAngle to angle
   endAngle (measured in degrees clockwise from due north) */

void SWFShape_drawArc(SWFShape shape, int r, float startAngle, float endAngle)
{
  int i;
  float controlx, controly, anchorx, anchory, x, y;

  /* first determine number of segments, 8 at most */
  int nSegs = 1 + floor(7*(endAngle-startAngle)/360);

  /* subangle is half the angle of each segment */
  float subangle = M_PI*(endAngle-startAngle)/nSegs/360;

  float angle = M_PI*startAngle/180;

  x = floor(r*sin(angle)+0.5);
  y = floor(r*cos(angle)+0.5);

  SWFShape_moveToRelative(shape, x, y);

  for(i=0; i<nSegs; ++i)
  {
    angle += subangle;
    controlx = r*sin(angle)/cos(subangle);
    controly = r*cos(angle)/cos(subangle);
    angle += subangle;
    anchorx = r*sin(angle);
    anchory = r*cos(angle);

    SWFShape_curveToRelative(shape,
			     (int)(floor(controlx+0.5)-x),
			     (int)(floor(controly+0.5)-y),
			     (int)(floor(anchorx+0.5)-x),
			     (int)(floor(anchory+0.5)-y));
    x = anchorx;
    y = anchory;
  }
}

#include "read.c"
#include <assert.h>
#define error(x) { printf("ERROR: %s\n",(x)); assert(0); }

/* yes, this is a total hack. */

void SWFShape_drawFontGlyph(SWFShape shape, SWFFont font, int c)
{
  byte *p = SWFFont_findCharacterGlyph(font, c);
  byte **f = &p;

  int moveBits, x, y;
  int straight, numBits;

  /* moveTos in the record are absolute, but we want to draw from the current
     location. grr. */

  int startX = SWFShape_getPenX(shape);
  int startY = SWFShape_getPenY(shape);

  byteAlign();

  if(readBits(f, 4) != 1) /* fill bits */
    error("SWFShape_drawCharacter: was expecting fill bits = 1");

  if(readBits(f, 4) != 0) /* line bits */
    error("SWFShape_drawCharacter: was expecting line bits = 0");

  /* now we get to parse the shape commands.  Oh boy. */

  /* the first one will be a non-edge block- grab the moveto loc */

  readBits(f, 6); /* type 0, etc. */

  moveBits = readBits(f, 5);
  x = readSBits(f, moveBits);
  y = readSBits(f, moveBits);

  SWFShape_moveTo(shape, x+startX, y+startY);

  if(readBits(f, 1) != 1) /* fill1 = 1 */
    error("SWFShape_drawCharacter says: oops.  Was expecting fill1 = 1.");

  /* we could just dump the rest if we had access to the shape's output buffer,
     but this isn't that hard.. */

  for(;;)
  {
    if(readBits(f, 1) == 0)
    {
      /* it's a moveTo or a shape end */

      if(readBits(f, 5) == 0)
	break;

      moveBits = readBits(f, 5);
      x = readSBits(f, moveBits);
      y = readSBits(f, moveBits);

      SWFShape_moveTo(shape, x+startX, y+startY);
      continue;
    }

    straight = readBits(f, 1);
    numBits = readBits(f, 4)+2;

    if(straight==1)
    {
      if(readBits(f, 1)) /* general line */
      {
	x = readSBits(f, numBits);
	y = readSBits(f, numBits);

	SWFShape_lineToRelative(shape, x, y);
      }
      else
	if(readBits(f, 1)) /* vert = 1 */
	  SWFShape_lineToRelative(shape, 0, readSBits(f, numBits));
	else
	  SWFShape_lineToRelative(shape, readSBits(f, numBits), 0);
    }
    else
    {
      int controlX = readSBits(f, numBits);
      int controlY = readSBits(f, numBits);
      int anchorX = readSBits(f, numBits);
      int anchorY = readSBits(f, numBits);

      SWFShape_curveToRelative(shape, controlX, controlY, anchorX, anchorY);
    }
  }

  /* no idea where the pen was left */
  SWFShape_moveTo(shape, startX, startY);
}
